/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - keyword-input.presentational.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  ElementRef,
  Inject,
  Injector,
  Input,
  OnInit,
  Output,
  ViewChild,
} from "@angular/core";
import {
  ControlValueAccessor,
  FormControl,
  NG_VALUE_ACCESSOR,
} from "@angular/forms";
import {
  MatChipInput,
  MatChipInputEvent,
  MatChipList,
} from "@angular/material/chips";
import {MatDialog} from "@angular/material/dialog";
import {
  FloatLabelType,
  MatFormFieldAppearance,
} from "@angular/material/form-field";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  distinctUntilChanged,
  tap,
} from "rxjs/operators";
import {KeywordPasteDialog} from "../../../core/components/dialogs/keyword-paste/keyword-paste.dialog";
import {AbstractInternalPresentational} from "../../../core/components/presentationals/abstract-internal/abstract-internal.presentational";
import {SOLIDIFY_CONSTANTS} from "../../../core/constants";
import {IconNamePartialEnum} from "../../../core/enums/partial/icon-name-partial.enum";
import {FormValidationHelper} from "../../../core/helpers/form-validation.helper";
import {LABEL_TRANSLATE} from "../../../core/injection-tokens/label-to-translate.injection-token";
import {KeyValue} from "../../../core/models/key-value.model";
import {
  isArray,
  isEmptyString,
  isNonEmptyString,
  isNotNullNorUndefined,
  isNullOrUndefined,
} from "../../../core/tools/is/is.tool";
import {ExtendEnum} from "../../../core/types/extend-enum.type";
import {ArrayUtil} from "../../../core/utils/array.util";
import {DialogUtil} from "../../../core/utils/dialog.util";
import {ObservableUtil} from "../../../core/utils/observable.util";
import {SsrUtil} from "../../../core/utils/ssr.util";
import {StringUtil} from "../../../core/utils/string.util";
import {LabelTranslateInterface} from "../../../label-translate-interface.model";

@Component({
  selector: "solidify-keyword-input",
  templateUrl: "./keyword-input.presentational.html",
  styleUrls: ["./keyword-input.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    {
      provide: NG_VALUE_ACCESSOR,
      multi: true,
      useExisting: KeywordInputPresentational,
    },
  ],
})

export class KeywordInputPresentational extends AbstractInternalPresentational implements OnInit, ControlValueAccessor {
  private readonly _KEY_ENTER: string = "Enter";

  @ViewChild(MatChipInput)
  inputChip: MatChipInput;

  @ViewChild("input")
  input: ElementRef;

  @Input()
  separatorKeys: string[] = [this._KEY_ENTER, SOLIDIFY_CONSTANTS.SPACE, SOLIDIFY_CONSTANTS.COMMA, SOLIDIFY_CONSTANTS.SEMICOLON, SOLIDIFY_CONSTANTS.SEPARATOR];

  @Input()
  placeholderToTranslate: string = this.labelTranslateInterface.inputKeywordInputLabel;

  @Input()
  readonly: boolean;

  @Input()
  formControl: FormControl;

  @Input()
  appearance: MatFormFieldAppearance = this.appearanceInputMaterial;

  @Input()
  positionLabel: FloatLabelType = this.positionLabelInputMaterial;

  @Input()
  hintToTranslate: string;

  @Input()
  capitalizeOnPaste: boolean = true;

  @Input()
  separatorToControlOnPaste: KeyValue[] = [
    {
      key: SOLIDIFY_CONSTANTS.COMMA,
      value: this.labelTranslate.inputComma,
    },
    {
      key: SOLIDIFY_CONSTANTS.SEMICOLON,
      value: this.labelTranslate.inputSemicolon,
    },
    {
      key: SOLIDIFY_CONSTANTS.UNDERSCORE,
      value: this.labelTranslate.inputUnderscore,
    },
    {
      key: SOLIDIFY_CONSTANTS.LINE_BREAK,
      value: this.labelTranslate.inputLineBreak,
    },
    {
      key: SOLIDIFY_CONSTANTS.DASH,
      value: this.labelTranslate.inputDash,
    },
    {
      key: "–",
      value: this.labelTranslate.inputDash,
    },
    {
      key: "—",
      value: this.labelTranslate.inputDash,
    },
    {
      key: "―",
      value: this.labelTranslate.inputDash,
    },
    {
      key: "‒",
      value: this.labelTranslate.inputDash,
    },
  ];

  @Input()
  extraInfoImage: ExtendEnum<IconNamePartialEnum> | undefined;

  @Input()
  extraInfoImageLabel: string | undefined;

  @Input()
  extraInfoImageLabelCallback: (value: string) => string = (value: string) => this.extraInfoImageLabel;

  clickExtraInfoImage(value: string, $event: Event): void {
    if (isNotNullNorUndefined($event)) {
      $event.stopPropagation();
    }
    this._extraInfoImageCallbackBS.next(value);
  }

  _spaceDelimiter: boolean = true;

  @Input()
  set spaceDelimiter(value: boolean) {
    this._spaceDelimiter = value;
    this._addOrRemoveSeparator(this._spaceDelimiter, SOLIDIFY_CONSTANTS.SPACE);
  }

  _enterDelimiter: boolean = true;

  @Input()
  set enterDelimiter(value: boolean) {
    this._enterDelimiter = value;
    this._addOrRemoveSeparator(this._enterDelimiter, this._KEY_ENTER);
  }

  @Input()
  showEmptyRequiredFieldInError: boolean = this.displayEmptyRequiredFieldInError;

  @Input()
  prefixIcon: ExtendEnum<IconNamePartialEnum>;

  private readonly _valueBS: BehaviorSubject<string[] | undefined> = new BehaviorSubject<string[] | undefined>(undefined);
  @Output("valueChange")
  readonly valueObs: Observable<string[] | undefined> = ObservableUtil.asObservable(this._valueBS);

  protected readonly _extraInfoImageCallbackBS: BehaviorSubject<string | undefined> = new BehaviorSubject<string | undefined>(undefined);
  @Output("extraInfoImageCallback")
  readonly extraInfoImageCallbackObs: Observable<string | undefined> = ObservableUtil.asObservable(this._extraInfoImageCallbackBS);

  get formValidationHelper(): typeof FormValidationHelper {
    return FormValidationHelper;
  }

  @ViewChild("chipList")
  chipList: MatChipList;

  constructor(private readonly _changeDetector: ChangeDetectorRef,
              protected readonly _injector: Injector,
              protected readonly _dialog: MatDialog,
              @Inject(LABEL_TRANSLATE) readonly labelTranslate: LabelTranslateInterface) {
    super(_injector);
  }

  ngOnInit(): void {
    if (isNotNullNorUndefined(this.formControl)) {
      this.subscribe(this.formControl.statusChanges.pipe(
        distinctUntilChanged(),
        tap(status => {
          this.chipList.errorState = status === SOLIDIFY_CONSTANTS.FORM_STATUS_INVALID;
          this.chipList._markAsTouched();
        }),
      ));
      this.subscribe(this.formControl.valueChanges.pipe(
        distinctUntilChanged(),
        tap(value => {
          this.propagateChange(value);
          this._valueBS.next(value);
          this.formControl.markAsPristine();
          this._changeDetector.detectChanges();
        }),
      ));
    }
  }

  propagateChange: (__: any) => void = (__: any) => {};

  registerOnChange(fn: any): void {
    this.propagateChange = fn;
  }

  registerOnTouched(fn: any): void {
  }

  setDisabledState(isDisabled: boolean): void {
  }

  writeValue(value: string[]): void {
  }

  remove(value: string): void {
    const list: string[] = this.formControl.value;
    const indexOf = list.indexOf(value);
    if (indexOf === -1) {
      return;
    }
    list.splice(indexOf, 1);
    this.formControl.markAsDirty();
    this.formControl.setValue(list);
  }

  add(event: MatChipInputEvent): void {
    const input = event.input;
    const value = (event.value || StringUtil.stringEmpty).trim();
    const list: string[] = this.formControl.value;

    if (!isEmptyString(value) && !list.includes(value)) {
      list.push(value.trim());
    }
    this.formControl.setValue(list);
    this.formControl.markAsDirty();

    if (input) {
      input.value = StringUtil.stringEmpty;
    }
  }

  getValue(): string[] {
    const value = this.formControl.value;
    return (!isNullOrUndefined(value) && isArray(value)) ? value.filter(v => v !== StringUtil.stringEmpty) : [];
  }

  private _addOrRemoveSeparator(withTheSeparator: boolean, separator: string): void {
    const indexOf = this.separatorKeys.indexOf(separator);
    if (withTheSeparator && indexOf === -1) {
      this.separatorKeys.push(separator);
    } else if (!withTheSeparator && indexOf !== -1) {
      this.separatorKeys.splice(indexOf, 1);
    }
  }

  paste($event: ClipboardEvent): void {
    const pastedString = ($event.clipboardData || SsrUtil.window?.clipboardData).getData("text");
    if (!isNonEmptyString(pastedString)) {
      return;
    }
    const listOfSeparatorFounded = [];
    this.separatorToControlOnPaste.forEach(separator => {
      if (pastedString.indexOf(separator.key) !== -1) {
        listOfSeparatorFounded.push(separator);
      }
    });

    let currentValue = this.formControl.value;
    if (!isArray(currentValue)) {
      currentValue = [];
    }

    if (listOfSeparatorFounded.length > 0) {
      this.subscribe(DialogUtil.open(this._dialog, KeywordPasteDialog, {
          stringPaste: pastedString,
          listSeparator: listOfSeparatorFounded,
        }, undefined, separator => {
          let listValue = pastedString.split(separator)
            .map((k: string) => StringUtil.capitalize(k.trim()))
            .filter(k => isNonEmptyString(k));

          listValue = ArrayUtil.distinct(listValue);

          this.formControl.setValue([...currentValue, ...listValue]);
          this.formControl.markAsDirty();
          $event.preventDefault();
          this._changeDetector.detectChanges();
        },
        () => {
          this.formControl.setValue([...currentValue, pastedString]);
          this.formControl.markAsDirty();
          this._changeDetector.detectChanges();
        }));
    } else {
      this.formControl.setValue([...currentValue, pastedString]);
      this.formControl.markAsDirty();
    }
    $event.preventDefault();
    $event.stopPropagation();
  }

  keydown($event: KeyboardEvent): void {
    if (this.separatorKeys.indexOf($event.key) !== -1) {
      this.inputChip.chipEnd.next({input: this.input.nativeElement, value: this.input.nativeElement.value} as MatChipInputEvent);
      $event.stopPropagation();
      $event.preventDefault();
    }
  }
}
